from itertools import product

import numpy as np
from scipy.integrate import simps as simpson

from .base import Distribution
from ..utils.constants import kB


class EnergyBarriers(Distribution):
    name = 'Energy barriers'
    parameter = [r'\tau_{0}', r'E_{m}', r'\Delta E']

    @staticmethod
    def distribution(tau, temperature, *args):
        t0, em, sigma = args
        m = np.exp(em / (kB * temperature)) * t0
        return temperature * np.exp(-0.5 * (np.log(tau/m)*kB*temperature/sigma)**2) / np.sqrt(2*np.pi)/sigma

    @staticmethod
    def rate(t0, e_a, te):
        return np.exp(-e_a / (kB * te)) / t0

    @staticmethod
    def energydistribution(e_a, mu, sigma):
        return np.exp(-0.5 * ((mu-e_a) / sigma) ** 2) / (np.sqrt(2 * np.pi) * sigma)

    @staticmethod
    def correlation(t, temperature, *args):
        tau0, e_m, e_b = args

        def integrand(e_a, ti, t0, mu, sigma, te):
            # correlation time would go to inf for higher energies, so we use rate
            return np.exp(-ti*EnergyBarriers.rate(t0, e_a, te)) * EnergyBarriers.energydistribution(e_a, mu, sigma)

        t = np.atleast_1d(t)
        temperature = np.atleast_1d(temperature)

        e_axis = np.linspace(max(0, e_m-50*e_b), e_m+50*e_b, num=5001)

        ret_val = np.array([simpson(integrand(e_axis, o, tau0, e_m, e_b, tt), e_axis)
                            for o in t for tt in temperature])

        return ret_val

    @staticmethod
    def susceptibility(omega, temperature, *args):
        # in contrast to other spectral densities, it's omega and temperature
        tau0, e_m, e_b = args

        def integrand_real(e_a, w, t0, mu, sigma, t):
            r = EnergyBarriers.rate(t0, e_a, t)
            return 1/(r**2 + w**2) * EnergyBarriers.energydistribution(e_a, mu, sigma)

        def integrand_imag(e_a, w, t0, mu, sigma, t):
            r = EnergyBarriers.rate(t0, e_a, t)
            return w*r/(r**2 + w**2) * EnergyBarriers.energydistribution(e_a, mu, sigma)

        omega = np.atleast_1d(omega)
        temperature = np.atleast_1d(temperature)

        e_axis = np.linspace(max(0, e_m-50*e_b), e_m+50*e_b, num=5001)
        ret_val = []
        for o, tt in product(omega, temperature):
            ret_val.append(simpson(integrand_real(e_axis, o, tau0, e_m, e_b, tt), e_axis) -
                           1j * simpson(integrand_imag(e_axis, o, tau0, e_m, e_b, tt), e_axis))

        return np.array(ret_val)

    @staticmethod
    def specdens(omega, temperature, *args):
        # in contrast to other spectral densities, it's omega and temperature
        tau0, e_m, e_b = args

        def integrand(e_a, w, t0, mu, sigma, t):
            r = EnergyBarriers.rate(t0, e_a, t)
            return r/(r**2 + w**2) * EnergyBarriers.energydistribution(e_a, mu, sigma)

        omega = np.atleast_1d(omega)
        temperature = np.atleast_1d(temperature)

        e_axis = np.linspace(max(0, e_m-50*e_b), e_m+50*e_b, num=5001)

        ret_val = np.array([simpson(integrand(e_axis, o, tau0, e_m, e_b, tt), e_axis)
                            for o in omega for tt in temperature])

        return ret_val

    @staticmethod
    def mean(*args):
        return args[1]*np.exp(args[2]/(kB*args[0]))

    @staticmethod
    def logmean(*args):
        return args[1] + args[2] / (kB * args[0])

    @staticmethod
    def max(*args):
        return args[1] * np.exp(args[2] / (kB * args[0]))
